use std::{collections::HashSet, hash::Hash};

use btleplug::{
    api::{Peripheral as _, PeripheralProperties},
    platform::{Peripheral, PeripheralId},
};
use serde::Serialize;
use uuid::Uuid;

use crate::playlist::PlayMode;

#[derive(Debug, Clone, Default, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct CentralState {
    pub is_ready: bool,
    pub is_scanning: bool,
    pub connected: Option<PeripheralId>,
}

#[derive(Debug, Clone, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct PeripheralItem {
    pub id: PeripheralId,
    pub address: String,
    pub represent: Option<String>,
    pub is_unknown: bool,
    pub is_connected: bool,
    pub rssi: Option<i16>,
    pub battery: Option<i16>,
    pub services: HashSet<PeripheralService>,
}

#[derive(Debug, Clone, Default, Eq, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct PeripheralService {
    pub uuid: Uuid,
    pub characteristics: HashSet<Uuid>,
}

#[derive(Debug, Clone, Default, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct ChannelState {
    pub is_playing: bool,
    pub strength: u32,
    pub strength_limit: u32,
    pub is_boosting: bool,
    pub boost_level: u32,
    pub boost_limit: u32,
    pub play_mode: PlayMode,
}

impl Hash for PeripheralService {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.uuid.hash(state);
    }
}

impl PartialEq for PeripheralService {
    fn eq(&self, other: &Self) -> bool {
        self.uuid == other.uuid
    }
}

impl PeripheralItem {
    pub async fn new(
        periperial: &Peripheral,
        properties: Option<&PeripheralProperties>,
        services: HashSet<PeripheralService>,
    ) -> Self {
        Self {
            id: periperial.id(),
            address: periperial.address().to_string(),
            represent: properties.and_then(|p| p.local_name.clone()),
            is_unknown: properties.and_then(|p| p.local_name.clone()).is_none(),
            is_connected: periperial.is_connected().await.unwrap_or(false),
            rssi: properties.and_then(|p| p.rssi),
            battery: properties.and_then(|p| p.tx_power_level),
            services,
        }
    }

    pub async fn new_with_cache(
        periperial: &Peripheral,
        properties: Option<&PeripheralProperties>,
        services: HashSet<Uuid>,
    ) -> Self {
        let service = services
            .into_iter()
            .map(|uuid| PeripheralService {
                uuid,
                ..Default::default()
            })
            .collect();
        Self {
            id: periperial.id(),
            address: periperial.address().to_string(),
            represent: properties.and_then(|p| p.local_name.clone()),
            is_unknown: properties.and_then(|p| p.local_name.clone()).is_none(),
            is_connected: periperial.is_connected().await.unwrap_or(false),
            rssi: properties.and_then(|p| p.rssi),
            battery: properties.and_then(|p| p.tx_power_level),
            services: service,
        }
    }

    pub async fn new_form_peripheral(
        periperial: &Peripheral,
        properties: Option<&PeripheralProperties>,
    ) -> Self {
        let services = periperial
            .services()
            .iter()
            .map(|s| {
                let characteristics = s.characteristics.iter().map(|c| c.uuid).collect();
                PeripheralService {
                    uuid: s.uuid,
                    characteristics,
                }
            })
            .collect();
        Self {
            id: periperial.id(),
            address: periperial.address().to_string(),
            represent: properties.and_then(|p| p.local_name.clone()),
            is_unknown: properties.and_then(|p| p.local_name.clone()).is_none(),
            is_connected: periperial.is_connected().await.unwrap_or(false),
            rssi: properties.and_then(|p| p.rssi),
            battery: properties.and_then(|p| p.tx_power_level),
            services,
        }
    }
}
